"""Functions that compare two columns.

Provides the foundation for phylogenetic inference tools.
"""

from hstrat._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub
