"""Classes to specify prior expectation for the distribution of MRCA generation
between hereditary stratigraphic columns, to assist in estimating the
phylogenetic relationship of hereditary stratigraphic columns."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub
